
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.Hashtable;

/**
 * Base class implementing all Enum features.
 * <P>
 * NOTE: This class requires JDK 1.2 / Java 2 to work properly.  It is recommended
 * that you compile it with the JDK 1.2 option -target 1.2 to prevent its accidental
 * use in Java 1.1 environments.
 * <P>
 * A subclass need only have a private ctor passing this class
 * the distinct id for each declared enum instance and the public static
 * declarations for the enums.
 * <P>
 * This class is not implemented to allow aliasing of multiple Enum instances to the same id value.
 *
 * @author R Evan Easton
 */
public abstract class Enum
	implements Serializable
{
	private final static Hashtable subclassInstanceDictionary__ = new Hashtable(5, 0.75f);
	private int id_;

	/**
	 * Construct an instance and index it in an instance dictionary that is separate for each subclass.
	 * @param id a distinct id to distinguish Enums from eachother per subclass.
	 * @exception DuplicateEnumIdException if another Enum has already been declared with the specified id for the concrete subclass
	 */
	protected Enum(int id)
		throws DuplicateEnumIdException
	{
		id_ = id;

		Hashtable instance_dictionary = (Hashtable)getInstanceDictionaryFor(getClass());
		synchronized(instance_dictionary)
		{
			Integer obj_id = new Integer(id);
			if(instance_dictionary.containsKey(obj_id))
				throw new DuplicateEnumIdException(getClass(), obj_id);

			instance_dictionary.put(obj_id, this);
		}		
	}

	/**
	 * Get the instance dictionary for the specified class.
	 * @return always returns a non-null reference
	 */
	private synchronized Hashtable getInstanceDictionaryFor(Class cls)
	{
		Hashtable ret = (Hashtable)subclassInstanceDictionary__.get(cls);
		if(ret == null)
		{
			ret = new Hashtable(5, 0.75f);
			subclassInstanceDictionary__.put(cls, ret);
		}
		return ret;
	}

	/**
	 * The unique object identity for this Enum instance within its subclass.
	 * <P><FONT COLOR="#cc0000">
	 * WARNING: Since it is public others can get the int value.  Therefore,
	 * this should be used only after careful consideration.
	 * </FONT></P>
	 */
	public final int asInt()
	{
		return id_;
	}

	/**
	 * Resolve the this instance to one of the unique instances in the instance dictionary.
	 * @exception NonExistentEnumException if the declared instance based on the id value cannot be found in the subclass's instance dictionary
	 */
	protected final Object readResolve()
		throws NonExistentEnumException
	{
		Hashtable instance_dictionary = (Hashtable)getInstanceDictionaryFor(getClass());

		// look up the already registered enum
		Enum ret = (Enum)instance_dictionary.get(new Integer(asInt()));

		// if it doesn't exist throw an exception
		if(ret == null)
			throw new NonExistentEnumException(getClass(), new Integer(asInt()));

		return ret;
	}

	/**
	 * The hashcode will be the hash code of simply concatenating the id and class name together.
	 * <P>
	 * It is marked final to prevent a subclass from breaking the method because of faulty assumptions.
	 */
	public final int hashCode()
	{
		String tmp = "" + asInt() + ":" + getClass().getName();
		return tmp.hashCode();
	}

	/**
	 * Value-wise comparison.
	 * This degenerates to reference equality because all instances
	 * are resolved to the declared instances even in deserialization.
	 * Try un-commenting the value compare code to see.
	 * <P>
	 * It is marked final to prevent a subclass from breaking the method because of faulty assumptions.
	 */
	public final boolean equals(Object obj)
	{
		boolean ret = (obj == this);

//		if(!ret && obj != null && obj.getClass() == getClass())
//			ret = ((Enum)obj).asInt() == asInt();

		return ret;
	}
}
