// Code for implementing and testing operator->* for smart pointers;
// supports pointers to members with up to two parameters.  Adding support
// for more parameters is straightforward.  Based on material presented at
// Software Development in May 1999 and appearing in a 1999 article in
// Dr. Dobbs Journal.
//
// I believe this file contains valid standard C++, but I can't find a
// compiler that agrees with me.  With minor tweaking it can be made to run
// under the following compilers:
//
// egcs 1.1.2: egcs doesn't seem to honor the using declaration in SP that
//   should make all the operator->* functions privately inherited from
//   SmartPtrBase public.  As a workaround, change the private inheritance
//   in line 1 below to public inheritance. (Search for "LINE 1".)
//
// Visual C++ 6.0 (SP2): VC6 doesn't like the using declaration inside
//   SP<T>.  However, it also seems to ignore the fact that members
//   inherited from a private base class are private to clients of the
//   derived class, so a workaround is to comment out Line 2 below.
//   (Search for "LINE 2".)
// 
// CodeWarrior Professional 4: CW4 fails to distinguish pointers to const
//   and non-const member function (it treats them as the same), so it
//   doesn't like the fact that SmartPtrBase declares operator->* templates
//   for both kinds of member pointers.  As a workaround, #define "CW4"
//   when compiling under CW4.  That will exclude the templates for pointers
//   to non-const member functions.  (Thanks for Andrei Alexandrescu for
//   the CW4 information and workaround.)
// 
// I'm told that C++ Builder 4 and Sun CC 5.0 reject the code for one
// reason or another, and I don't know of workarounds.  (I also have no
// access to these compilers, so I can't try to find workarounds.)  If you
// test this on other compilers, please let me know what you find.
//
// Written by Scott Meyers, April-May 1999.  Send comments to
// smeyers@aristeia.com.

#include <iostream>
#include <utility>

namespace ScottMeyers {
  using std::pair;
  using std::make_pair;

  template <typename ObjectType,      // class with the mem func to be called
            typename ReturnType,      // the return type of the mem func
            typename MemFuncPtrType>  // the type of the ptr-to-mem-func
  class PMFC {                        // PMFC = "Pending Member Function Call"
  public:
    typedef pair<ObjectType*, MemFuncPtrType> CallInfo;

    PMFC(const CallInfo& info)
      : callInfo(info) {}

    // Pending call via operator->* to a member function taking 0 parameters
    ReturnType operator()() const 
      { return (callInfo.first->*callInfo.second)(); }

    // Pending call via operator->* to a member function taking 1 parameter
    template <typename Param1Type>
    ReturnType operator()(Param1Type p1) const
      { return (callInfo.first->*callInfo.second)(p1); }

    // Pending call via operator->* to a member function taking 2 parameters
    template <typename Param1Type,
              typename Param2Type>
    ReturnType operator()(Param1Type p1, Param2Type p2) const
      { return (callInfo.first->*callInfo.second)(p1, p2); }

  private:
    CallInfo callInfo;
  };
  
  template <typename T>          // base class for smart pointers wishing
  class SmartPtrBase {           // to support operator->*
  public:
    SmartPtrBase(T *initVal): ptr(initVal) {}

#ifndef CW4
    // for use with a non-const member function taking 0 parameters
	template <typename ReturnType>
	const PMFC<T, ReturnType, ReturnType (T::*)()> 
	operator->*(ReturnType (T::*pmf)()) const
      { return make_pair(ptr, pmf); }
#endif

    // for use with a const member function taking 0 parameters
	template <typename ReturnType>
	const PMFC<T, ReturnType, ReturnType (T::*)() const> 
	operator->*(ReturnType (T::*pmf)() const) const
      { return make_pair(ptr, pmf); }

#ifndef CW4
    // for use with a non-const member function taking 1 parameter
	template <typename ReturnType,
              typename Param1Type>
	const PMFC<T, ReturnType, ReturnType (T::*)(Param1Type)>
	operator->*(ReturnType (T::*pmf)(Param1Type)) const
      { return make_pair(ptr, pmf); }
#endif

    // for use with a const member function taking 1 parameter
	template <typename ReturnType,
              typename Param1Type>
	const PMFC<T, ReturnType, ReturnType (T::*)(Param1Type) const>
	operator->*(ReturnType (T::*pmf)(Param1Type) const) const
      { return make_pair(ptr, pmf); }

#ifndef CW4
    // for use with a non-const member function taking 2 parameters
	template <typename ReturnType,
              typename Param1Type,
              typename Param2Type>
	const PMFC<T, ReturnType, ReturnType (T::*)(Param1Type, Param2Type)>
	operator->*(ReturnType (T::*pmf)(Param1Type, Param2Type)) const
      { return make_pair(ptr, pmf); }
#endif

    // for use with a const member function taking 2 parameters
	template <typename ReturnType,
              typename Param1Type,
              typename Param2Type>
	const PMFC<T, ReturnType, ReturnType (T::*)(Param1Type, Param2Type) const>
	operator->*(ReturnType (T::*pmf)(Param1Type, Param2Type) const) const
      { return make_pair(ptr, pmf); }

  protected:
    T *ptr;
  };

} // namespace ScottMeyers


// polluting the global namespace is gauche, so we put our test stuff in
// an unnamed namespace
namespace {
  using std::cout;
  using std::endl;

  template<typename T>
  class SP: public ScottMeyers::SmartPtrBase<T> {   // LINE 1 (see above)
  public:
    SP(T *initVal): ScottMeyers::SmartPtrBase<T>(initVal) {}

    using ScottMeyers::SmartPtrBase<T>::operator->*;   // LINE 2 (see above)
                                          // make public the operator->*
                                          // functions in the private
                                          // base class 

    // Other smart pointer functions (e.g., operator*, operator->, etc.)
    // should also be in this class.  Since this file tests only 
    // operator->* functionality, I've omitted these other functions.

  };

  // I got lazy and made all the functions below inline.  Also, I didn't
  // use void return values, because VC6 won't compile the "return"
  // statements in the PMFC templates.  This is a compiler limitation,
  // because the C++ standard makes clear that the code should compile.  In
  // other words, the PMFC and operator->* templates *should* work with
  // pointers to member functions that have a void return type.
  class Widget {
  public:
    int nonconst0Params()
      { cout << "nonconst0Params" << endl;  return 0; }
    int const0Params() const
      { cout << "const0Params" << endl; return 0; }

    double nonconst1Param(int x) 
      { cout << "nonconst1Param, x = " << x << endl;  return x; }
    double const1Param(int x) const
      { cout << "const1Param, x = " << x << endl;  return x; }

    const char* nonconst2Params(int x, bool b) 
      { 
        cout << "nonconst2Params, x = " << x << ", b = " << b << endl; 
        return "Implementing operator->* is fun!";
      }
    const char* const2Params(int x, bool b) const
      { 
        cout << "const2Params, x = " << x << ", b = " << b << endl; 
        return "Implementing operator->* is fun!";
      }
  };

} // unnamed namespace


int main()
{
  using std::cout;
  using std::endl;

  // In what follows, PMF is always a typedef for a pointer to a non-const
  // member function, and PCMF is always a typedef for a pointer to a const
  // member function

  SP<Widget> pw = new Widget;

  // test calls to 0-parameter member functions
  {
    typedef int (Widget::*PMF)();
    PMF p1 = &Widget::nonconst0Params;
    int r1 = (pw->*p1)();
    cout << "Result = " << r1 << endl << endl;

    typedef int (Widget::*PCMF)() const;
    PCMF p2 = &Widget::const0Params;
    int r2 = (pw->*p2)();
    cout << "Result = " << r2 << endl << endl;
  }
    
  // test calls to 1-parameter member functions
  {
    typedef double (Widget::*PMF)(int);
    PMF p1 = &Widget::nonconst1Param;
    double r1 = (pw->*p1)(1);
    cout << "Result = " << r1 << endl << endl;

    typedef double (Widget::*PCMF)(int) const;
    PCMF p2 = &Widget::const1Param;
    double r2 = (pw->*p2)(1);
    cout << "Result = " << r2 << endl << endl;
  }
    
  // test calls to 2-parameter member functions
  {
    typedef const char* (Widget::*PMF)(int, bool);
    PMF p1 = &Widget::nonconst2Params;
    const char* r1 = (pw->*p1)(2, true);
    cout << "Result = " << r1 << endl << endl;

    typedef const char* (Widget::*PCMF)(int, bool) const;
    PCMF p2 = &Widget::const2Params;
    const char* r2 = (pw->*p2)(2, false);
    cout << "Result = " << r2 << endl << endl;
  }
    
  return 0;
}
