// Code for implementing and testing operator->* for smart pointers; uses
// partial specialization; supports pointers to members with up to two
// parameters.  Adding support for more parameters is straightforward.
// Based on material appearing in a 1999 article in Dr. Dobbs Journal.
//
// This file will be easier to understand after you've looked at the
// companion file that implements the same behavior without using partial
// specialization.
//
// This file compiles under egcs 1.1.2.  It does not compile under Visual
// C++ 6, because that compiler lacks support for partial specialization.
// I'm told that Sun CC 5.0 also rejects it.
// If you test it on other compilers, please let me know what you find out.
//
// Written by Scott Meyers with considerable help from Mark Rodgers, May
// 1999.  Send comments to smeyers@aristeia.com.

#include <iostream>
#include <utility>

namespace ScottMeyers {
  using std::pair;

  template <typename T>                         // traits class
    struct MemFuncTraits {};

  template <typename R, typename O>             // partial 
    struct MemFuncTraits<R (O::*)()> {          // specialization for
    typedef R ReturnType;                       // zero-parameter non-const
    typedef O ObjectType;                       // member funcs
  };

  template <typename R, typename O>             // partial 
    struct MemFuncTraits<R (O::*)() const> {    // specialization for
    typedef R ReturnType;                       // zero-parameter const
    typedef O ObjectType;                       // member funcs
  };

  template <typename R, typename O, 
            typename P1>                        // partial
    struct MemFuncTraits<R (O::*)(P1)> {        // specialization for
    typedef R ReturnType;                       // one-parameter non-const
    typedef O ObjectType;                       // member funcs
  };

  template <typename R, typename O, 
            typename P1>                        // partial
    struct MemFuncTraits<R (O::*)(P1) const> {  // specialization for
    typedef R ReturnType;                       // one-parameter const
    typedef O ObjectType;                       // member funcs
  };

  template <typename R, typename O, 
            typename P1, typename P2>           // partial
    struct MemFuncTraits<R (O::*)(P1,P2)> {     // specialization for
    typedef R ReturnType;                       // two-parameter non-const
    typedef O ObjectType;                       // member funcs
  };

  template <typename R, typename O, 
            typename P1, typename P2>              // partial
    struct MemFuncTraits<R (O::*)(P1,P2) const> {  // specialization for
    typedef R ReturnType;                          // two-parameter const
    typedef O ObjectType;                          // member funcs
  };


  template <typename MemFuncPtrType>  // the type of the ptr-to-mem-func
  class PMFC {                        // PMFC = "Pending Member Function Call"
  public:
    typedef typename MemFuncTraits<MemFuncPtrType>::ObjectType ObjectType;
    typedef typename MemFuncTraits<MemFuncPtrType>::ReturnType ReturnType;

    typedef pair<ObjectType*, MemFuncPtrType> CallInfo;

    PMFC(const CallInfo& info)
      : callInfo(info) {}

    // Pending call via operator->* to a member function taking 0 parameters
    ReturnType operator()() const 
      { return (callInfo.first->*callInfo.second)(); }

    // Pending call via operator->* to a member function taking 1 parameter
    template <typename Param1Type>
    ReturnType operator()(Param1Type p1) const
      { return (callInfo.first->*callInfo.second)(p1); }

    // Pending call via operator->* to a member function taking 2 parameters
    template <typename Param1Type,
              typename Param2Type>
    ReturnType operator()(Param1Type p1, Param2Type p2) const
      { return (callInfo.first->*callInfo.second)(p1, p2); }

  private:
    CallInfo callInfo;
  };

  // In contrast to the approach that doesn't use partial specialization,
  // there is no SmartPtrBase template, because the only code to be shared
  // in smart pointer classes is the single member template for
  // operator->*.  That template can easily become part of the smart
  // pointer boilerplate (like operator* and operator->); it's included in
  // the SP template below.
  
} // namespace ScottMeyers


// polluting the global namespace is gauche, so we put our test stuff in
// an unnamed namespace
namespace {
  using ScottMeyers::PMFC;
  using std::make_pair;
  using std::cout;
  using std::endl;

  template<typename T>
  class SP {
  public:
    SP(T *initVal): ptr(initVal) {}

    template <typename MemFuncPtrType>
	const PMFC<MemFuncPtrType>
    operator->*(MemFuncPtrType pmf) const
      { return make_pair(ptr, pmf); }

    // Other smart pointer functions (e.g., operator*, operator->, etc.)
    // should also be in this class.  Since this file tests only 
    // operator->* functionality, I've omitted these other functions.

  private:
    T *ptr;
  };

  class Widget {
  public:
    int nonconst0Params()
      { cout << "nonconst0Params" << endl;  return 0; }
    int const0Params() const
      { cout << "const0Params" << endl; return 0; }

    double nonconst1Param(int x) 
      { cout << "nonconst1Param, x = " << x << endl;  return x; }
    double const1Param(int x) const
      { cout << "const1Param, x = " << x << endl;  return x; }

    const char* nonconst2Params(int x, bool b) 
      { 
        cout << "nonconst2Params, x = " << x << ", b = " << b << endl; 
        return "Implementing operator->* is fun!";
      }
    const char* const2Params(int x, bool b) const
      { 
        cout << "const2Params, x = " << x << ", b = " << b << endl; 
        return "Implementing operator->* is fun!";
      }
  };

} // unnamed namespace


int main()
{
  using std::cout;
  using std::endl;

  // In what follows, PMF is always a typedef for a pointer to a non-const
  // member function, and PCMF is always a typedef for a pointer to a const
  // member function

  SP<Widget> pw = new Widget;

  // test calls to 0-parameter member functions
  {
    typedef int (Widget::*PMF)();
    PMF p1 = &Widget::nonconst0Params;
    int r1 = (pw->*p1)();
    cout << "Result = " << r1 << endl << endl;

    typedef int (Widget::*PCMF)() const;
    PCMF p2 = &Widget::const0Params;
    int r2 = (pw->*p2)();
    cout << "Result = " << r2 << endl << endl;
  }
    
  // test calls to 1-parameter member functions
  {
    typedef double (Widget::*PMF)(int);
    PMF p1 = &Widget::nonconst1Param;
    double r1 = (pw->*p1)(1);
    cout << "Result = " << r1 << endl << endl;

    typedef double (Widget::*PCMF)(int) const;
    PCMF p2 = &Widget::const1Param;
    double r2 = (pw->*p2)(1);
    cout << "Result = " << r2 << endl << endl;
  }
    
  // test calls to 2-parameter member functions
  {
    typedef const char* (Widget::*PMF)(int, bool);
    PMF p1 = &Widget::nonconst2Params;
    const char* r1 = (pw->*p1)(2, true);
    cout << "Result = " << r1 << endl << endl;

    typedef const char* (Widget::*PCMF)(int, bool) const;
    PCMF p2 = &Widget::const2Params;
    const char* r2 = (pw->*p2)(2, false);
    cout << "Result = " << r2 << endl << endl;
  }
    
  return 0;
}
